"""Parameter context for python protocols."""

from typing import List, Optional, Union, Dict

from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.parameters import parameter_definition, validation
from opentrons.protocols.parameters.types import (
    ParameterChoice,
    ParameterDefinitionError,
)
from opentrons.protocol_engine.types import RunTimeParameter, RunTimeParamValuesType

from ._parameters import Parameters

_ParameterDefinitionTypes = Union[
    parameter_definition.ParameterDefinition[int],
    parameter_definition.ParameterDefinition[bool],
    parameter_definition.ParameterDefinition[float],
    parameter_definition.ParameterDefinition[str],
]


class ParameterContext:
    """Public context for adding parameters to a protocol."""

    def __init__(self, api_version: APIVersion) -> None:
        """Initializes a parameter context for user-set parameters."""
        self._api_version = api_version
        self._parameters: Dict[str, _ParameterDefinitionTypes] = {}

    def add_int(
        self,
        display_name: str,
        variable_name: str,
        default: int,
        minimum: Optional[int] = None,
        maximum: Optional[int] = None,
        choices: Optional[List[ParameterChoice]] = None,
        description: Optional[str] = None,
        unit: Optional[str] = None,
    ) -> None:
        """Creates an integer parameter, settable within a given range or list of choices.

        Arguments:
            display_name: The display name of the int parameter as it would show up on the frontend.
            variable_name: The variable name the int parameter will be referred to in the run context.
            default: The default value the int parameter will be set to. This will be used in initial analysis.
            minimum: The minimum value the int parameter can be set to (inclusive). Mutually exclusive with choices.
            maximum: The maximum value the int parameter can be set to (inclusive). Mutually exclusive with choices.
            choices: A list of possible choices that this parameter can be set to.
                Mutually exclusive with minimum and maximum.
            description: A description of the parameter as it will show up on the frontend.
            unit: An optional unit to be appended to the end of the integer as it shown on the frontend.
        """
        validation.validate_variable_name_unique(variable_name, set(self._parameters))
        parameter = parameter_definition.create_int_parameter(
            display_name=display_name,
            variable_name=variable_name,
            default=default,
            minimum=minimum,
            maximum=maximum,
            choices=choices,
            description=description,
            unit=unit,
        )
        self._parameters[parameter.variable_name] = parameter

    def add_float(
        self,
        display_name: str,
        variable_name: str,
        default: float,
        minimum: Optional[float] = None,
        maximum: Optional[float] = None,
        choices: Optional[List[ParameterChoice]] = None,
        description: Optional[str] = None,
        unit: Optional[str] = None,
    ) -> None:
        """Creates a float parameter, settable within a given range or list of choices.

        Arguments:
            display_name: The display name of the float parameter as it would show up on the frontend.
            variable_name: The variable name the float parameter will be referred to in the run context.
            default: The default value the float parameter will be set to. This will be used in initial analysis.
            minimum: The minimum value the float parameter can be set to (inclusive). Mutually exclusive with choices.
            maximum: The maximum value the float parameter can be set to (inclusive). Mutually exclusive with choices.
            choices: A list of possible choices that this parameter can be set to.
                Mutually exclusive with minimum and maximum.
            description: A description of the parameter as it will show up on the frontend.
            unit: An optional unit to be appended to the end of the float as it shown on the frontend.
        """
        validation.validate_variable_name_unique(variable_name, set(self._parameters))
        parameter = parameter_definition.create_float_parameter(
            display_name=display_name,
            variable_name=variable_name,
            default=validation.ensure_float_value(default),
            minimum=validation.ensure_optional_float_value(minimum),
            maximum=validation.ensure_optional_float_value(maximum),
            choices=validation.ensure_float_choices(choices),
            description=description,
            unit=unit,
        )
        self._parameters[parameter.variable_name] = parameter

    def add_bool(
        self,
        display_name: str,
        variable_name: str,
        default: bool,
        description: Optional[str] = None,
    ) -> None:
        """Creates a boolean parameter with allowable values of "On" (True) or "Off" (False).

        Arguments:
            display_name: The display name of the boolean parameter as it would show up on the frontend.
            variable_name: The variable name the boolean parameter will be referred to in the run context.
            default: The default value the boolean parameter will be set to. This will be used in initial analysis.
            description: A description of the parameter as it will show up on the frontend.
        """
        validation.validate_variable_name_unique(variable_name, set(self._parameters))
        parameter = parameter_definition.create_bool_parameter(
            display_name=display_name,
            variable_name=variable_name,
            default=default,
            choices=[
                {"display_name": "On", "value": True},
                {"display_name": "Off", "value": False},
            ],
            description=description,
        )
        self._parameters[parameter.variable_name] = parameter

    def add_str(
        self,
        display_name: str,
        variable_name: str,
        default: str,
        choices: Optional[List[ParameterChoice]] = None,
        description: Optional[str] = None,
    ) -> None:
        """Creates a string parameter, settable among given choices.

        Arguments:
            display_name: The display name of the string parameter as it would show up on the frontend.
            variable_name: The variable name the string parameter will be referred to in the run context.
            default: The default value the string parameter will be set to. This will be used in initial analysis.
            choices: A list of possible choices that this parameter can be set to.
                Mutually exclusive with minimum and maximum.
            description: A description of the parameter as it will show up on the frontend.
        """
        validation.validate_variable_name_unique(variable_name, set(self._parameters))
        parameter = parameter_definition.create_str_parameter(
            display_name=display_name,
            variable_name=variable_name,
            default=default,
            choices=choices,
            description=description,
        )
        self._parameters[parameter.variable_name] = parameter

    def set_parameters(self, parameter_overrides: RunTimeParamValuesType) -> None:
        """Sets parameters to values given by client, validating them as well.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        for variable_name, override_value in parameter_overrides.items():
            try:
                parameter = self._parameters[variable_name]
            except KeyError:
                raise ParameterDefinitionError(
                    f"Parameter {variable_name} is not defined as a parameter for this protocol."
                )
            validated_value = validation.ensure_value_type(
                override_value, parameter.parameter_type
            )
            parameter.value = validated_value

    def export_parameters_for_analysis(self) -> List[RunTimeParameter]:
        """Exports all parameters into a protocol engine models for reporting in analysis.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return [
            parameter.as_protocol_engine_type()
            for parameter in self._parameters.values()
        ]

    def export_parameters_for_protocol(self) -> Parameters:
        """Exports all parameters into a protocol run usable parameters object.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return Parameters(
            parameters={
                parameter.variable_name: parameter.value
                for parameter in self._parameters.values()
            }
        )



"""Protocol engine state module."""

from .state import State, StateStore, StateView
from .state_summary import StateSummary
from .config import Config
from .commands import (
    CommandState,
    CommandView,
    CommandSlice,
    CurrentCommand,
)
from .command_history import CommandEntry
from .labware import LabwareState, LabwareView
from .pipettes import PipetteState, PipetteView, HardwarePipette
from .modules import ModuleState, ModuleView, HardwareModule
from .module_substates import (
    MagneticModuleId,
    MagneticModuleSubState,
    HeaterShakerModuleId,
    HeaterShakerModuleSubState,
    TemperatureModuleId,
    TemperatureModuleSubState,
    ThermocyclerModuleId,
    ThermocyclerModuleSubState,
    ModuleSubStateType,
)
from .geometry import GeometryView
from .motion import MotionView, PipetteLocationData

__all__ = [
    # top level state value and interfaces
    "State",
    "StateStore",
    "StateView",
    "StateSummary",
    # static engine configuration
    "Config",
    # command state and values
    "CommandState",
    "CommandView",
    "CommandSlice",
    "CurrentCommand",
    "CommandEntry",
    # labware state and values
    "LabwareState",
    "LabwareView",
    # pipette state and values
    "PipetteState",
    "PipetteView",
    "HardwarePipette",
    # module state and values
    "ModuleState",
    "ModuleView",
    "HardwareModule",
    "MagneticModuleId",
    "MagneticModuleSubState",
    "HeaterShakerModuleId",
    "HeaterShakerModuleSubState",
    "TemperatureModuleId",
    "TemperatureModuleSubState",
    "ThermocyclerModuleId",
    "ThermocyclerModuleSubState",
    "ModuleSubStateType",
    # computed geometry state
    "GeometryView",
    # computed motion state
    "MotionView",
    "PipetteLocationData",
]



"""Wrappers for the legacy, Protocol API v2 execution pipeline."""
import asyncio
from typing import Dict, Iterable, Optional, cast

from anyio import to_thread

from opentrons_shared_data.labware.dev_types import (
    LabwareDefinition as LegacyLabwareDefinition,
)
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.calibration_storage.helpers import uri_from_details
from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.modules.types import (
    ModuleModel as LegacyModuleModel,
    TemperatureModuleModel as LegacyTemperatureModuleModel,
    MagneticModuleModel as LegacyMagneticModuleModel,
    ThermocyclerModuleModel as LegacyThermocyclerModuleModel,
    HeaterShakerModuleModel as LegacyHeaterShakerModuleModel,
)
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_engine import ProtocolEngine
from opentrons.protocol_engine.types import RunTimeParamValuesType
from opentrons.protocol_reader import ProtocolSource, ProtocolFileRole
from opentrons.util.broker import Broker

from opentrons.protocol_api import (
    ProtocolContext as LegacyProtocolContext,
    InstrumentContext as LegacyPipetteContext,
    ModuleContext as LegacyModuleContext,
    Labware as LegacyLabware,
    Well as LegacyWell,
    ParameterContext,
    create_protocol_context,
)
from opentrons.protocol_api.core.engine import ENGINE_CORE_API_VERSION
from opentrons.protocol_api.core.legacy.load_info import (
    LoadInfo as LegacyLoadInfo,
    InstrumentLoadInfo as LegacyInstrumentLoadInfo,
    LabwareLoadInfo as LegacyLabwareLoadInfo,
    ModuleLoadInfo as LegacyModuleLoadInfo,
)

from opentrons.protocols.parse import PythonParseMode, parse
from opentrons.protocols.execution.execute import run_protocol
from opentrons.protocols.types import (
    Protocol as LegacyProtocol,
    JsonProtocol as LegacyJsonProtocol,
    PythonProtocol as LegacyPythonProtocol,
)

# The earliest Python Protocol API version ("apiLevel") where the protocol's simulation
# and execution will be handled by Protocol Engine, rather than the legacy machinery.
#
# Note that even when simulation and execution are handled by the legacy machinery,
# Protocol Engine still has some involvement for analyzing the simulation and
# monitoring the execution.
LEGACY_PYTHON_API_VERSION_CUTOFF = ENGINE_CORE_API_VERSION


# The earliest JSON protocol schema version where the protocol is executed directly by
# Protocol Engine, rather than going through Python Protocol API v2.
LEGACY_JSON_SCHEMA_VERSION_CUTOFF = 6


class LegacyFileReader:
    """Interface to read Protocol API v2 protocols prior to execution."""

    @staticmethod
    def read(
        protocol_source: ProtocolSource,
        labware_definitions: Iterable[LabwareDefinition],
        python_parse_mode: PythonParseMode,
    ) -> LegacyProtocol:
        """Read a PAPIv2 protocol into a data structure."""
        protocol_file_path = protocol_source.main_file
        protocol_contents = protocol_file_path.read_text(encoding="utf-8")
        legacy_labware_definitions: Dict[str, LegacyLabwareDefinition] = {
            uri_from_details(
                namespace=lw.namespace,
                load_name=lw.parameters.loadName,
                version=lw.version,
            ): cast(LegacyLabwareDefinition, lw.dict(exclude_none=True))
            for lw in labware_definitions
        }
        data_file_paths = [
            data_file.path
            for data_file in protocol_source.files
            if data_file.role == ProtocolFileRole.DATA
        ]

        return parse(
            protocol_file=protocol_contents,
            filename=protocol_file_path.name,
            extra_labware=legacy_labware_definitions,
            extra_data={
                data_path.name: data_path.read_bytes() for data_path in data_file_paths
            },
            python_parse_mode=python_parse_mode,
        )


# TODO (spp, 2023-04-05): Remove 'legacy' wording since this is the context we are using
#  for all python protocols.
class LegacyContextCreator:
    """Interface to construct Protocol API v2 contexts."""

    _USE_SIMULATING_CORE = False

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        protocol_engine: ProtocolEngine,
    ) -> None:
        """Prepare the LegacyContextCreator.

        Args:
            hardware_api: The hardware control interface.
                Will be wrapped in a `SynchronousAdapter`.
                May be real hardware or a simulator.
            protocol_engine: Interface for the context to load labware offsets.
        """
        self._hardware_api = hardware_api
        self._protocol_engine = protocol_engine

    def create(
        self,
        protocol: LegacyProtocol,
        broker: Optional[LegacyBroker],
        equipment_broker: Optional[Broker[LegacyLoadInfo]],
    ) -> LegacyProtocolContext:
        """Create a Protocol API v2 context."""
        extra_labware = (
            protocol.extra_labware
            if isinstance(protocol, LegacyPythonProtocol)
            else None
        )

        bundled_data = (
            protocol.bundled_data
            if isinstance(protocol, LegacyPythonProtocol)
            else None
        )

        return create_protocol_context(
            api_version=protocol.api_level,
            hardware_api=self._hardware_api,
            deck_type=self._protocol_engine.state_view.config.deck_type.value,
            protocol_engine=self._protocol_engine,
            protocol_engine_loop=asyncio.get_running_loop(),
            broker=broker,
            equipment_broker=equipment_broker,
            extra_labware=extra_labware,
            use_simulating_core=self._USE_SIMULATING_CORE,
            bundled_data=bundled_data,
        )


class LegacySimulatingContextCreator(LegacyContextCreator):
    """Interface to construct PAPIv2 contexts using simulating implementations.

    Avoids some calls to the hardware API for performance.
    See `opentrons.protocols.context.simulator`.
    """

    _USE_SIMULATING_CORE = True


class LegacyExecutor:
    """Interface to execute Protocol API v2 protocols in a child thread."""

    @staticmethod
    async def execute(
        protocol: LegacyProtocol,
        context: LegacyProtocolContext,
        parameter_context: Optional[ParameterContext],
        run_time_param_values: Optional[RunTimeParamValuesType],
    ) -> None:
        """Execute a PAPIv2 protocol with a given ProtocolContext in a child thread."""
        await to_thread.run_sync(
            run_protocol, protocol, context, parameter_context, run_time_param_values
        )


__all__ = [
    # Re-exports of user-facing Python Protocol APIv2 stuff:
    # TODO(mc, 2022-08-22): remove, no longer "legacy", so re-exports unnecessary
    "LegacyProtocolContext",
    "LegacyLabware",
    "LegacyWell",
    "LegacyPipetteContext",
    "LegacyModuleContext",
    # Re-exports of internal stuff:
    "LegacyProtocol",
    "LegacyJsonProtocol",
    "LegacyPythonProtocol",
    "LegacyLoadInfo",
    "LegacyLabwareLoadInfo",
    "LegacyInstrumentLoadInfo",
    "LegacyModuleLoadInfo",
    "LegacyModuleModel",
    "LegacyMagneticModuleModel",
    "LegacyTemperatureModuleModel",
    "LegacyThermocyclerModuleModel",
    "LegacyHeaterShakerModuleModel",
    # legacy typed dicts
    "LegacyLabwareDefinition",
]
